#! /usr/bin/env python3

# NEED to pass descriptions from the source task, not Terra Task Name

import json
import logging
from typing import Literal, Dict, List, Tuple, Union, Any
from copy import deepcopy
from collections import defaultdict, namedtuple

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def phb_row_to_string(
    row,
) -> str:
    """
    Convert a dictionary row to a tab-separated string.

    Args:
        row (Dict): The dictionary row to convert.
    Returns:
        The tab-separated string representation of the row.
    """
    # use json.dumps to get double-quote nested strings when within arrays/maps
    return "\t".join(str(v) for v in row.values())


def phb_resolve_classifications(
    var_name: str,
    description: str,
) -> Literal["docker", "database", "reference", "runtime", "general"]:
    """
    Resolve classifications based on variable name and descriptions.

    Args:
        var_name (str): The variable name.
        description (str): The variable description.
    Returns:
        The string classification category.
    """
    return (
        "docker"
        if any(_ in var_name for _ in ["docker", "docker_image"])
        and "internal component" not in description.lower()
        else (
            "database"
            if any(_ in var_name for _ in ["db", "database"])
            and "internal component" not in description.lower()
            else (
                "reference"
                if any(_ in var_name for _ in ["_ref", "ref_", "reference"])
                and "internal component" not in description.lower()
                else (
                    "runtime"
                    if any(
                        _ in var_name
                        for _ in ["cpu", "disk", "disk_size", "_mem", "mem_", "memory"]
                    )
                    and "internal component" not in description.lower()
                    else "general"
                )
            )
        )
    )


def phb_resolve_defaults(
    default: Union[str, Dict[str, Any]],
) -> str:
    """
    Resolve complex default values, including handling of functions like sub, addition, and nested functions like basename.

    Args:
        default (str | dict): The default value to resolve.
            - If str: returned as-is.
            - If dict: expected structure { "func": ..., "args": [...] }.
    Returns:
        The resolved default value as a string.
    """

    # recursively parse nested basename functions
    def parse_basename_func(default):
        if isinstance(default, dict) and default.get("func") == "basename":
            base, exts = parse_basename_func(default["args"][0][0])
            exts += [default["args"][1][0]]
            return base, exts
        else:
            return default, []

    # basename function
    if isinstance(default, dict) and default.get("func") == "basename":
        base, exts = parse_basename_func(default)
        default = f"basename of {base} (without {', '.join(exts)})"

    # substitution function
    elif isinstance(default, dict) and default.get("func") == "sub":
        default = f"{default['args'][0][0]} where '{default['args'][1][0]}' is substituted with '{default['args'][2][0]}'"

    # addition function
    elif isinstance(default, dict) and default.get("func") == "_add":
        default = "".join([_[0] for _ in default["args"]])

    # use json.dumps to double-quote strings
    elif isinstance(default, dict) or isinstance(default, list):
        default = json.dumps(default)

    # Note converting all default values into strings for consistency between dict and input tsv
    else:
        default = str(default) if default != None else ""

    return default


def phb_resolve_var_name(
    var_name: str,
) -> Union[Tuple[str, str], Tuple[None, None]]:
    """
    Resolve an input variable name and its associated workflow from the dict file

    Args:
        var_name (str): The input variable name from the dict file.
    Returns:
        Tuple[str, str]: A tuple containing the workflow name and the variable name (in that order).
    """
    parts = var_name.split(".")
    # variable is defined at the main workflow level
    if len(parts) == 1:
        return None, var_name
    # variable is defined at the task/subworkflow level
    elif len(parts) <= 2:
        return parts[0], parts[1]
    return None, None


def phb_resolve_descriptions(var_name: str) -> str:
    """
    Resolve descriptions based on variable.

    Args:
        var_name (str): The variable name.
    Returns:
        The resolved description.
    """
    common_descriptions = {
        ("cpu",): "Number of CPUs to allocate to the task",
        ("disk", "disk_size"): "Amount of storage (in GB) to allocate to the task",
        ("mem", "memory"): "Amount of memory/RAM (in GB) to allocate to the task",
        ("docker", "docker_image"): "The Docker container to use for the task",
        (
            "timezone",
        ): "Set the time zone to get an accurate date of analysis (uses UTC by default)",
    }
    description = "DESCRIPTION"
    for keys, desc in common_descriptions.items():
        for k in keys:
            if k in var_name:
                description = desc
                break
    return description


def phb_parse_tsv(tsv_file: str) -> Tuple[Dict[str, Dict[str, str]], List[str]]:
    """
    Read input TSV file and return its contents.

    Args:
        tsv_file (str): The path to the TSV file.
    Returns:
        A dictionary mapping workflow display names to lists of dictionaries, where each inner dictionary represents an input variable and its associated metadata.
        A list of the file header
    """
    logger.debug(f"Parsing TSV file: {tsv_file}")
    with open(tsv_file, "r", newline="") as f:
        rows = f.readlines()
        tsv_file_header = rows[0].strip().split("\t")

        tsv_dict = {}
        for row in rows[1:]:
            row = row.strip().split("\t")
            tsv_row = dict(zip(tsv_file_header, row))

            # Only keep columns that exist in this row
            tsv_row = {k: v for k, v in tsv_row.items() if k in tsv_file_header}
            try:
                workflow_list = [x for x in tsv_row["Workflow"].split(", ") if x]
            except KeyError:
                raise KeyError(f"Incorrectly formatted row: {tsv_row}")

            # Populate tsv_dict with all workflows that use this input variable
            for wf in workflow_list:
                if wf not in tsv_dict:
                    tsv_dict[wf] = []
                tsv_dict[wf].append(tsv_row)

    return tsv_dict, tsv_file_header


def phb_generate_rows(combined_dict: dict) -> List[Dict[str, str]]:
    """
    Extract rows from a synthesized dictionary of TSV and BioBlueprint I/O metadata
    """
    all_rows_prep = []
    # prepare independent rows for any contrasting metadata
    for context_vars, wf_dict in combined_dict.items():
        t_rows = defaultdict(list)
        for wf, metadata in wf_dict.items():
            t_row = [
                (
                    k,
                    v,
                )
                for k, v in metadata.items()
            ] + list(context_vars)
            sorted_t_row = tuple(sorted(t_row, key=lambda x: x[0]))
            t_rows[sorted_t_row].append(wf)
        for t_row, wfs in t_rows.items():
            t_row_dict = {k: v for k, v in t_row}
            wfs_str = ", ".join(sorted(wfs))
            t_row_dict["Workflow"] = wfs_str
            all_rows_prep.append(t_row_dict)

    all_rows_sorted = [
        y
        for y in sorted(
            all_rows_prep,
            key=lambda x: (
                x.get("Terra Status", "").lower()
                != "required",  # False values are sorted before True values in python
                x.get("Terra Task Name", "").lower(),
                x.get("Variable", "").lower(),
            ),
        )
    ]

    return all_rows_sorted


def phb_write_tsv(
    metadata_dict: dict,
    tsv_file_header: list,
    output_file: str = "all_inputs_sorted.tsv",
) -> None:
    """
    Write a dictionary to a TSV file.
    Expected structure: { "Workflow Name": [ { "Terra Task Name": ..., "Variable": ..., ...}, ... ], ... }
    Args:
        output_file (str): The path to the output TSV file.
    """

    all_rows_sorted = phb_generate_rows(metadata_dict)

    with open(output_file, "w") as f:
        f.write("\t".join(tsv_file_header) + "\n")

        # unpack into a dictionary
        for row in all_rows_sorted:
            try:
                finalized_row = {k: row[k] for k in tsv_file_header}
            except KeyError:
                raise KeyError(f"{row} missing key")
            print(phb_row_to_string(finalized_row), file=f)


def phb_parse_io_dict(
    io_dict: Dict[str, Dict],
    wf_alias_map: Dict[str, str],
) -> Tuple[Dict[str, Dict[str, str]], Dict[str, Dict[str, str]]]:
    """
    Parse the I/O dictionary of all inputs/outputs with their associated workflows.

    Args:
        io_dict (str): The I/O dictionary generated from Bioblueprint workflow parsing.
        wf_alias_map (dict): A dictionary mapping workflow file paths to workflow display names.

    Returns:
        A dictionary mapping workflow display names to lists of dictionaries, where each inner dictionary represents an input variable and its associated metadata.
        A dictionary mapping workflow display names to lists of dictionaries, where each inner dictionary represents an output variable and its associated metadata.
    """
    input_dict = {}
    output_dict = {}
    logger.debug(f"Parsing IO Dictionary")
    for wf_wdl_name, info in io_dict.items():
        path = info["path"]
        inputs = info["inputs"]
        outputs = info["outputs"]

        # If the workflow is not hosted on dockstore or no longer used, set the display name to nothing. Will be removed later.
        wf_display_name = wf_alias_map.get(path, "")
        if not wf_display_name:
            continue
        input_dict[wf_display_name] = []
        output_dict[wf_display_name] = []

        # Parse inputs first
        logger.debug(
            f"Parsing inputs: {wf_wdl_name} with path: {path} and display name: {wf_display_name}"
        )
        for var_name, attribute in inputs.items():
            # Terra only considers the first 3 levels of workflow inputs, skip if variable is too deeply nested
            if len(var_name.split(".")) > 2:
                continue

            # variable is defined at the task/subworkflow level
            terra_task_name, var_name = phb_resolve_var_name(var_name)

            # variable is defined at the main workflow level
            if terra_task_name is None:
                terra_task_name = wf_wdl_name

            var_type = attribute["type"].split("?")[0]
            description = phb_resolve_descriptions(var_name)
            default = phb_resolve_defaults(attribute["default"])
            status = (
                "Required"
                if ((attribute["default"] == None) and ("?" not in attribute["type"]))
                else "Optional"
            )
            classification = phb_resolve_classifications(var_name, description)

            input_dict[wf_display_name].append(
                {
                    "Terra Task Name": terra_task_name,
                    "Variable": var_name,
                    "Type": var_type,
                    "Description": description,
                    "Default Value": default,
                    "Terra Status": status,
                    "Classification": classification,
                    "Manual Override": "",
                }
            )

        # Parse outputs
        logger.debug(
            f"Parsing outputs: {wf_wdl_name} with path: {path} and display name: {wf_display_name}"
        )
        for var_name, var_type in outputs.items():
            output_dict[wf_display_name].append(
                {
                    "Variable": var_name,
                    "Type": var_type.split("?")[0],
                }
            )

    return input_dict, output_dict


def phb_prep_tsv_dict(in_dict: dict) -> Dict[str, Dict[str, Dict[str, str]]]:
    """
    Prepare a dictionary to compare with another for row-row compatibility

    Args:
        input_dict (dict): An parsed I/O dict obtained from an input TSV or I/O dictionary made from BioBlueprint

    Returns:
        output_dict (dict): A dictionary with Terra Task Name and Variable Name extracted for comparisons.
    """
    contextual_variables = {"Terra Task Name", "Variable"}
    output_dict = defaultdict(dict)
    for wf, rows in in_dict.items():
        # unpack a row to obtain the variables we want, and those to maintain
        for row in rows:
            row_key = tuple(
                (
                    k,
                    v,
                )
                for k, v in row.items()
                if k in contextual_variables
            )
            if row.get("Manual Override"):
                output_dict[row_key][wf] = {
                    k: v for k, v in row.items() if k not in contextual_variables
                }
            else:
                output_dict[row_key][wf] = {"Description": row["Description"]}
    return output_dict


def phb_prep_io_dict(in_dict: dict) -> Dict[str, Dict[str, Dict[str, str]]]:
    """
    Prepare a dictionary to compare with another for row-row compatibility

    Args:
        input_dict (dict): An parsed I/O dict obtained from an input TSV or I/O dictionary made from BioBlueprint

    Returns:
        output_dict (dict): A dictionary with Terra Task Name and Variable Name extracted for comparisons.
    """
    contextual_variables = {"Terra Task Name", "Variable"}
    output_dict = defaultdict(dict)
    for wf, rows in in_dict.items():
        # unpack a row to obtain the variables we want, and those to maintain
        for row in rows:
            row_key = tuple(
                (
                    k,
                    v,
                )
                for k, v in row.items()
                if k in contextual_variables
            )
            output_dict[row_key][wf] = {
                k: v for k, v in row.items() if k not in contextual_variables
            }
    return output_dict


def phb_update_io(
    i_or_o_dict: dict,
    tsv_dict: dict,
    internal_component_str="Internal component, do not modify",
) -> Tuple[Dict, List, List]:
    """
    Update the input variables in the I/O tsv file based on differences found in the dict input dictionary.

    Args:
        i_or_o_dict (dict): From the input dict. Dictionary mapping workflow display names to a list of dictionaries containing input variables and their metadata.
        tsv_dict (dict): From the IO input tsv file. Dictionary mapping workflow display names to a list of dictionaries containing input variables and their metadata.

    Returns:
        dict: The updated dictionary of valid inputs for all workflows.
    """
    # Flatten both input dictionaries into lists of tuples for easier comparison
    # Ignore wf entries that are most likely subworkflows. We only want to report/document the ones that show up in Terra.
    parsed_io_dict = phb_prep_io_dict(i_or_o_dict)
    tsv_descrip_dict = phb_prep_tsv_dict(tsv_dict)

    logger.debug(
        f"Checking for exact matches between {len(i_or_o_dict)} BioBlueprint rows and {len(tsv_descrip_dict)} TSV rows"
    )
    additions = []
    for dict_row_key, wf_dict in parsed_io_dict.items():
        for dict_wf_name, metadata in wf_dict.items():
            # does this I/O exist in the tsv currently
            if dict_row_key in tsv_descrip_dict:
                # if the workflow isn't there, grab a description for it, otherwise use the existing
                if dict_wf_name not in tsv_descrip_dict[dict_row_key]:
                    descriptions = set(
                        v["Description"]
                        for v in tsv_descrip_dict[dict_row_key].values()
                    )
                    # use the first description
                    if descriptions:
                        # check for multiple descriptions, omit internal components
                        if len(descriptions) > 1:
                            descriptions = sorted(
                                x for x in descriptions if x != internal_component_str
                            )
                        description = list(descriptions)[0]
                    else:
                        description = "DESCRIPTION"
                    additions.append([dict_wf_name, dict_row_key, metadata])
                else:
                    description = tsv_descrip_dict[dict_row_key][dict_wf_name][
                        "Description"
                    ]
            # this entry does not exist
            else:
                # set an empty description
                description = ""
                additions.append([dict_wf_name, dict_row_key, metadata])
            if "Default Value" in metadata:
                if (
                    internal_component_str == description
                    or metadata["Default Value"] == ""
                ):
                    metadata["Classification"] = "general"
            metadata["Description"] = description
            # only preexisting entries can be manually overridden
            if dict_wf_name in tsv_descrip_dict[dict_row_key]:
                if not tsv_descrip_dict[dict_row_key][dict_wf_name].get(
                    "Manual Override"
                ):
                    tsv_descrip_dict[dict_row_key][dict_wf_name] = metadata

    # remove extraneous entries
    removals = []
    for dict_row_key, wf_dict in tsv_descrip_dict.items():
        if dict_row_key not in parsed_io_dict:
            for dict_wf_name, metadata in wf_dict.items():
                removals.append([dict_wf_name, dict_row_key, metadata])
        else:
            for dict_wf_name, metadata in wf_dict.items():
                if dict_wf_name not in parsed_io_dict[dict_row_key]:
                    removals.append([dict_wf_name, dict_row_key, metadata])
    for dict_wf_name, dict_row_key, metadata in removals:
        del tsv_descrip_dict[dict_row_key][dict_wf_name]

    return tsv_descrip_dict, additions, removals


def phb_generate_report(
    additions: list,
    removals: list,
    output_file: str = "io_changelog.tsv",
) -> None:
    """
    Generate a report of changes made to all I/O variables.
    """
    logger.debug(f"Generating report of changes to I/O variables: {output_file}")
    with open(output_file, "w", newline="") as f:

        print("ADDITIONS:", file=f)
        for wf_name, row, description in additions:
            row_dict = {k: v for k, v in row}
            row_dict["Description"] = description
            print(f"{wf_name}: {row_dict}", file=f)

        print(
            "\n---------------------------------------------------------------------------------\n",
            file=f,
        )
        print("REMOVALS:", file=f)
        for wf_name, row, description in removals:
            row_dict = {k: v for k, v in row}
            row_dict["Description"] = description
            print(f"{wf_name}: {row_dict}", file=f)


def main(
    io_dict: str,
    input_tsv_file: str,
    output_tsv_file: str,
    wf_alias_map: Dict,
    out_dir: str,
    out_prefix: str,
) -> None:
    input_dict, output_dict = phb_parse_io_dict(io_dict, wf_alias_map)

    # Parse and update input variables first
    logger.info(f"Updating {input_tsv_file}")
    tsv_input_dict, input_file_header = phb_parse_tsv(input_tsv_file)
    updated_input_dict, input_adds, input_rms = phb_update_io(
        input_dict, tsv_input_dict
    )
    logger.info(f"{len(input_adds)} additions; {len(input_rms)} removals")

    # Create the updated input tsv file and generate a report of changes made to input variables
    phb_generate_report(
        input_adds, input_rms, output_file=f"{out_dir}{out_prefix}_inputs_changelog.tsv"
    )
    phb_write_tsv(
        updated_input_dict,
        input_file_header,
        output_file=f"{out_dir}{out_prefix}_inputs.tsv",
    )

    # Parse and update output variables next
    logger.info(f"Updating {output_tsv_file}")
    tsv_output_dict, output_file_header = phb_parse_tsv(output_tsv_file)
    updated_output_dict, output_adds, output_rms = phb_update_io(
        output_dict, tsv_output_dict
    )
    logger.info(f"{len(output_adds)} additions; {len(output_rms)} removals")

    # Create the updated output tsv file and generate a report of changes made to output variables
    phb_generate_report(
        output_adds,
        output_rms,
        output_file=f"{out_dir}{out_prefix}_outputs_changelog.tsv",
    )
    phb_write_tsv(
        updated_output_dict,
        output_file_header,
        output_file=f"{out_dir}{out_prefix}_outputs.tsv",
    )



#! /usr/bin/env python3

import re
import yaml
import logging
import requests
from typing import Dict, Any

# NEED to modify populate PR/create new function to enable editing existing PR
"""Repository interface and Repository-specific functions"""

logger = logging.getLogger(__name__)


def read_yaml(
    yaml_file: str,
) -> Dict[str, Any]:
    """
    Read a YAML file and return its contents.

    Args:
        yaml_file (str): The path to the YAML file.
    Returns:
        The contents of the YAML file as a dictionary.
    """
    logger.debug(f"Reading YAML file: {yaml_file}")
    with open(yaml_file, "r") as f:
        return yaml.safe_load(f)


def phb_get_wf_display_name(dockstore_yaml_file: str) -> Dict[str, str]:
    """
    Parse the dockstore yaml to create a mapping of workflow file paths to display names.

    Args:
        dockstore_yaml_file (str): The path to the dockstore yaml file.
    Returns:
        A dictionary mapping workflow file paths to workflow display names.
    """
    logger.debug(f"Parsing Dockstore YAML file: {dockstore_yaml_file}")

    wf_alias_map = {}  # { "file_path" : "Workflow/Display Name"}

    # Structure of dockstore_yaml dictionary:
    # { 'version': ..., 'workflows': [ { 'name': ..., 'subclass': ..., 'primaryDescriptorPath': ..., 'testParameterFiles': [...] }, ... ]}
    dockstore_yaml = read_yaml(dockstore_yaml_file)

    for wf in dockstore_yaml["workflows"]:
        # this next part should be removed and the dockstore yaml should be updated so we don't have to do this.
        # or the "Workflow" column in `all_inputs.tsv` should match the display names in the dockstore yaml.
        display_name = re.sub(r"_PHB$", "", wf["name"])
        fp = wf["primaryDescriptorPath"][1:]  # removes leading "/"
        if display_name == "NCBI-AMRFinderPlus":
            display_name = "AMRFinderPlus"
        if display_name == "Kraken2_ONT":
            display_name = "Kraken_ONT"
        if display_name == "Kraken2_PE":
            display_name = "Kraken_PE"
        if display_name == "Kraken2_SE":
            display_name = "Kraken_SE"
        if display_name == "Gambit_Core":
            continue

        wf_alias_map[fp] = display_name
    return wf_alias_map


def download_pr(owner, repo, pr_number):
    """Download a PR from GitHub and return its body text."""

    url = f"https://api.github.com/repos/{owner}/{repo}/pulls/{pr_number}"
    response = requests.get(url)
    if response.status_code == 200:
        pr_data = response.json()
        return pr_data.get("body", "")
    else:
        raise Exception(f"Failed to fetch PR: {response.status_code} - {response.text}")


def define_phb_standards(phb_dir):
    """Define standardized paths for github.com/theiagen/public_health_bioinformatics"""
    pr_template_path = phb_dir + ".github/PULL_REQUEST_TEMPLATE/code.md"
    pr_pop_func = populate_phb_pr
    i_tsv_path = f"{phb_dir}docs/assets/tables/all_inputs.tsv"
    o_tsv_path = f"{phb_dir}docs/assets/tables/all_outputs.tsv"
    dockstore_path = f"{phb_dir}.dockstore.yml"
    return pr_template_path, pr_pop_func, i_tsv_path, o_tsv_path, dockstore_path


def populate_phb_pr(wf_dict, task_dict, io_diff_dict, testing_dict, pr_template):
    """Populate the PR template with the modified workflows, tasks, inputs, and outputs."""

    # populate the expected section headings
    wf_task_header = "## :zap: Impacted Workflows/Tasks"
    impact_section = "This PR may lead to different results in pre-existing outputs:"
    inputs_header = "### ➡️ Inputs"
    outputs_header = "### ⬅️ Outputs"
    testing_header = "## :test_tube: Testing"
    scenarios_header = "### Suggested Scenarios for Reviewer to Test"

    # initialize the PR and populate from last to first sections to preserve indices
    pr_list = pr_template.splitlines()

    # add the testing workflows that don't already exist
    test_index = pr_list.index(testing_header)
    entered_tests = set()
    if scenarios_header in pr_list:
        scenarios_index = pr_list.index(scenarios_header)
        for line in pr_list[test_index + 1 : scenarios_index]:
            if line.strip().startswith(("- [ ]", "- [x]")):
                if line.strip().startswith("- [ ]"):
                    wf_name_prep = line.split("- [ ]")[1].strip()
                    complete = False
                else:
                    wf_name_prep = line.split("- [x]")[1].strip()
                    complete = True
                wf_name_prep1 = re.sub(
                    r"\[([^\]]+)\]\([^)]+\)", r"\1", wf_name_prep
                )  # remove markdown links
                wf_data = wf_name_prep1.split(" ")  # remove any comments after the name
                wf_name = wf_data[0].replace("~~", "")  # replace strike-outs
                if len(wf_data) > 1:
                    comment = " ".join(wf_data[1:])
                else:
                    comment = ""
                if wf_name in testing_dict:
                    testing_dict[wf_name]["complete"] = complete
                    if comment:
                        testing_dict[wf_name]["comment"] = comment
                    logger.debug(
                        f"{wf_name} testing complete: {complete}, comment: {comment}"
                    )
                else:
                    testing_dict[wf_name] = {"complete": complete, "comment": comment}
                entered_tests.add(wf_name)

    test_str = ""
    for x in {k: v for k, v in sorted(testing_dict.items(), key=lambda x: x[0])}:
        if x not in entered_tests:
            test_str += f"- [ ] {x}\n"
            testing_dict[x]["complete"] = False  # add to dict
    if test_str:
        pr_list.insert(test_index + 1, test_str)

    # add the input/output changes
    # use the key to indicate the workflow/task
    in_str, out_str = "\n", "\n"
    for k, v in io_diff_dict.items():
        if v["added_inputs"] or v["removed_inputs"]:
            in_str += f"\n<details>\n<summary>{k}</summary>\n\n"
            if v["added_inputs"]:
                in_str += "\n".join(f"\t+ {x}" for x in v["added_inputs"]) + "\n"
            if v["removed_inputs"]:
                in_str += "\n".join(f"\t- {x}" for x in v["removed_inputs"]) + "\n"
            in_str += "\n</details>\n"
        if v["added_outputs"] or v["removed_outputs"]:
            out_str += f"\n<details>\n<summary>{k}</summary>\n\n"
            if v["added_outputs"]:
                out_str += "\n".join(f"\t+ {x}" for x in v["added_outputs"]) + "\n"
            if v["removed_outputs"]:
                out_str += "\n".join(f"\t- {x}" for x in v["removed_outputs"]) + "\n"
            out_str += "\n</details>\n"

    # remove I/O entries prior to population
    in_index = pr_list.index(inputs_header)
    out_index = pr_list.index(outputs_header)
    for index in reversed(range(out_index + 1, test_index)):
        pr_list.pop(index)
    pr_list.insert(out_index + 1, out_str)
    for index in reversed(range(in_index + 1, out_index)):
        pr_list.pop(index)
    pr_list.insert(in_index + 1, in_str)

    # add the workflow/task changes
    wf_str = ""
    # use the key to indicate the type of change
    for k, v in wf_dict.items():
        if v:
            wf_str += "\n\n".join(f"{k} `{x}`" for x in v) + "\n\n"
    task_str = ""
    for k, v in task_dict.items():
        if v:
            task_str += "\n\n".join(f"{k} `{x}`" for x in v) + "\n\n"
    wf_task_index = pr_list.index(wf_task_header)
    # remove preexisting entries

    impact_index = [
        i for i, line in enumerate(pr_list) if line.strip().startswith(impact_section)
    ][0]
    for index in reversed(range(wf_task_index + 1, impact_index)):
        pr_list.pop(index)

    # readd a line break, then add workflows before tasks
    pr_list.insert(wf_task_index + 1, "\n")
    if task_str:
        pr_list.insert(wf_task_index + 1, f"\n### Tasks\n{task_str}")
    if wf_str:
        pr_list.insert(wf_task_index + 1, f"\n### Workflows\n{wf_str}\n")

    return "\n".join(pr_list), testing_dict



#! /usr/bin/env python3

# NEED to make it so graph nodes are specific to a file + task/workflow, not a task/workflow name
# will require modifying wdl.py during graph construction to identify the path of the nodes
# then reexpose the attribute errors that account for duplicate names/remove altogether
# NEED to parse diffs to identify the task that was changed, not using the whole file

# BEYOND MVP
# NEED to incorporate conditional parsing to link I/O modifications to more refined testing paths

import re
import os
import sys
import json
import argparse
import logging
from pathlib import Path
from datetime import datetime
from bioblueprint.lib.wdl import compile_repo, extract_wdl_nodes
from bioblueprint.lib.graph import obtain_graph_paths
from git import Repo
from bioblueprint.lib.repos import (
    download_pr,
    define_phb_standards,
    phb_get_wf_display_name,
)
from bioblueprint.tools.update_io import main as phb_update_io

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def prep_output_dir(out_dir=None):
    """Prepare an output directory"""
    if out_dir:
        clean_dir = str(Path(out_dir).resolve())
    else:
        cur_time = datetime.now().strftime("%Y%m%d")
        clean_dir = f"{os.getcwd()}/bioblueprint_{cur_time}/"
    if not os.path.isdir(clean_dir):
        os.mkdir(clean_dir)
    return clean_dir


def prep_output_prefix(input_path, suffix):
    """Prepare the output base path"""
    out_base = input_path
    if input_path.endswith("/"):
        out_base = out_base[:-1]
    out_base = os.path.basename(re.sub(suffix + r"$", "", out_base))
    return out_base


def initialize_bioblueprint():
    parser = argparse.ArgumentParser(
        description="BioBlueprint workflow dependency compiler and automation"
    )

    in_par = parser.add_argument_group("Input parameters")
    in_par.add_argument(
        "-i", "--input", help="Local Git repository directory", required=True
    )
    in_par.add_argument(
        "-px", "--prefix", help='Workflow prefix; DEFAULT: "wf_"', default="wf_"
    )
    in_par.add_argument(
        "-sx", "--suffix", help='Language suffix; DEFAULT: ".wdl"', default=".wdl"
    )

    git_par = parser.add_argument_group("Git parameters")
    git_par.add_argument(
        "-d", "--derived_branch", help="Derived branch name; DEFAULT: active branch"
    )
    git_par.add_argument(
        "-s",
        "--source_branch",
        help='Source branch name; DEFAULT: "main"',
        default="main",
    )

    repo_par = parser.add_argument_group("Repository parameters")
    repo_par.add_argument(
        "-o",
        "--owner",
        help="GitHub repository owner; DEFAULT: theiagen",
        default="theiagen",
    )
    repo_par.add_argument(
        "-r",
        "--repository",
        help="GitHub repository name; DEFAULT: basename of -i/--input",
    )
    repo_par.add_argument(
        "-pr",
        "--pull_request",
        type=int,
        help="Modify existing pull request number; DEFAULT: create new PR",
    )

    run_par = parser.add_argument_group("Runtime parameters")
    run_par.add_argument(
        "-v", "--verbose", action="store_true", help="Enable verbose debug output"
    )

    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    args.input = str(Path(args.input).resolve())
    if not args.input.endswith("/"):
        args.input += "/"
    out_dir = prep_output_dir()
    out_base = prep_output_prefix(args.input, args.suffix)
    if not args.repository:
        args.repository = os.path.basename(args.input[:-1])

    logger.warning(
        "BioBlueprint may change your branch. Ensure you are pulled and committed appropriately"
    )

    return args, out_dir, out_base


def compile_branch(repo, branch_name=""):
    if branch_name:
        logger.info(f"Checking out branch: {branch_name}")
        repo.git.checkout(branch_name)
    else:
        branch_name = repo.active_branch.name
        logger.info(f"On branch: {branch_name}")
    commit = repo.head.commit.tree
    return commit, branch_name


def check_merge_status(repo, derived_name, source_name):
    """Check if source branch is merged into derived branch"""
    commits_ahead = repo.iter_commits(f"{derived_name}..{source_name}")
    ahead_count = sum(1 for _ in commits_ahead)
    if ahead_count > 0:
        return False
    return True


def identify_changed_files(src_commit, der_commit, suffix):
    """Identify changed files between two commits"""
    # get the diff
    diff = src_commit.diff(der_commit)
    added_files = [
        x.b_path for x in diff.iter_change_type("A") if x.b_path.endswith(suffix)
    ]
    modified_files = [
        x.b_path for x in diff.iter_change_type("M") if x.b_path.endswith(suffix)
    ]
    removed_files = [
        x.b_path for x in diff.iter_change_type("D") if x.b_path.endswith(suffix)
    ]

    logger.info(f"Added files: {added_files}")
    logger.info(f"Modified files: {modified_files}")
    logger.info(f"Removed files: {removed_files}")
    return added_files, modified_files, removed_files


def parse_branches(repo, derived_branch, source_branch, suffix):
    src_commit, source_branch = compile_branch(repo, branch_name=source_branch)
    der_commit, derived_branch = compile_branch(repo, branch_name=derived_branch)

    if source_branch == derived_branch:
        logger.warning(
            f"Source branch {source_branch} is the same as derived branch {derived_branch}"
        )

    added_files, modified_files, removed_files = identify_changed_files(
        src_commit, der_commit, suffix
    )
    if not (added_files or modified_files or removed_files):
        logger.warning(
            "No added, modified, or removed WDL files detected between source and derived branches"
        )

    return derived_branch, source_branch, added_files, modified_files, removed_files


def compile_affected_nodes(input_dir, files, prefix):
    """Compile nodes that are affected by modifications in specific workflow files"""
    nodes_dict = {"wf": [], "task": []}
    for file_ in files:
        node_names = extract_wdl_nodes(input_dir + file_)
        logger.debug(f"Parsed WDL file {file_}: {node_names}")
        if os.path.basename(file_).startswith(prefix):
            nodes_dict["wf"].extend(node_names)
        else:
            nodes_dict["task"].extend(node_names)
    return nodes_dict


def parse_io_modifications(
    der_io_dict, src_io_dict, input_exposure_depth=1, output_exposure_depth=0
):
    """Identify differences in I/O and report by its parent"""
    diff_dict = {}
    for wf in der_io_dict:
        if wf in src_io_dict:
            mod_inputs = sorted(
                set(der_io_dict[wf]["inputs"]) - set(src_io_dict[wf]["inputs"])
            )
            mod_outputs = sorted(
                set(der_io_dict[wf]["outputs"]) - set(src_io_dict[wf]["outputs"])
            )
            rm_inputs = sorted(
                set(src_io_dict[wf]["inputs"]) - set(der_io_dict[wf]["inputs"])
            )
            rm_outputs = sorted(
                set(src_io_dict[wf]["outputs"]) - set(der_io_dict[wf]["outputs"])
            )
            # only output exposed inputs / outputs
            terra_mod_inputs = [
                x for x in mod_inputs if x.count(".") <= input_exposure_depth
            ]
            terra_mod_outputs = [
                x for x in mod_outputs if x.count(".") <= output_exposure_depth
            ]
            terra_rm_inputs = [
                x for x in rm_inputs if x.count(".") <= input_exposure_depth
            ]
            terra_rm_outputs = [
                x for x in rm_outputs if x.count(".") <= output_exposure_depth
            ]
            if mod_inputs or mod_outputs or rm_inputs or rm_outputs:
                diff_dict[wf] = {
                    "added_inputs": terra_mod_inputs,
                    "added_outputs": terra_mod_outputs,
                    "removed_inputs": terra_rm_inputs,
                    "removed_outputs": terra_rm_outputs,
                }
        else:
            diff_dict[wf] = {
                "added_inputs": sorted(
                    [
                        x
                        for x in der_io_dict[wf]["inputs"]
                        if x.count(".") <= input_exposure_depth
                    ]
                ),
                "added_outputs": sorted(
                    [
                        x
                        for x in der_io_dict[wf]["outputs"]
                        if x.count(".") <= output_exposure_depth
                    ]
                ),
                "removed_inputs": [],
                "removed_outputs": [],
            }
    return diff_dict


def parse_wf_task_modifications(new_nodes, perturbed_nodes, removed_nodes):
    """Identify differences in workflows and tasks"""
    wf_diff_dict = {"\\+": [], "Δ": [], "\\-": []}
    task_diff_dict = {"\\+": [], "Δ": [], "\\-": []}
    for node in new_nodes["wf"]:
        wf_diff_dict["\\+"].append(node[1])
    for node in new_nodes["task"]:
        task_diff_dict["\\+"].append(node[1])
    for node in perturbed_nodes["wf"]:
        wf_diff_dict["Δ"].append(node[1])
    for node in perturbed_nodes["task"]:
        task_diff_dict["Δ"].append(node[1])
    for node in removed_nodes["wf"]:
        wf_diff_dict["\\-"].append(node[1])
    for node in removed_nodes["task"]:
        task_diff_dict["\\-"].append(node[1])
    return {k: sorted(v) for k, v in wf_diff_dict.items()}, {
        k: sorted(v) for k, v in task_diff_dict.items()
    }


def translate_tags_and_names(io_dict_der, io_dict_src, wf_path2name={}):
    """Translate tag hashes between derived and source references based on shared PATHs"""
    path2tag_der = {v["path"]: k for k, v in io_dict_der.items()}
    path2tag_src = {v["path"]: k for k, v in io_dict_src.items()}
    tag2name_der = {
        v: wf_path2name[k] for k, v in path2tag_der.items() if k in wf_path2name
    }
    tag2name_src = {
        v: wf_path2name[k] for k, v in path2tag_src.items() if k in wf_path2name
    }
    src2der = {v: path2tag_der[k] for k, v in path2tag_src.items() if k in path2tag_der}
    return src2der, tag2name_der, tag2name_src


def main():
    args, out_dir, out_base = initialize_bioblueprint()

    # initialize the repository and git metadata
    if not os.path.isdir(args.input + "/.git"):
        raise SystemExit(
            f"Input directory {args.input} is not the head of a git repository"
        )
    repo = Repo(args.input)

    starting_branch = repo.active_branch.name
    logger.info(f"Initial branch: {starting_branch}")
    logger.info(f"Checking out branch: {args.derived_branch}")
    repo.git.checkout(args.derived_branch)

    logger.info(f"Checking branch congruence (merge status)")
    merged = check_merge_status(repo, args.derived_branch, args.source_branch)
    if not merged:
        logger.error(
            f"Source branch {args.source_branch} is not merged into derived branch {args.derived_branch}."
        )
        raise SystemExit(
            f"Please merge {args.source_branch} into {args.derived_branch} before proceeding."
        )

    derived_branch, source_branch, added_files, modified_files, removed_files = (
        parse_branches(repo, args.derived_branch, args.source_branch, args.suffix)
    )

    # initialize repo-specific functions and metadata
    io_path = f"{out_dir}{out_base}.io.json"
    pr_out_path = f"{out_dir}{out_base}.pr.md"
    if args.repository == "public_health_bioinformatics":
        # define the standardized paths for PHB and required functions
        pr_template_path, pr_pop_func, i_tsv_path, o_tsv_path, dockstore_path = (
            define_phb_standards(args.input)
        )
        io_update_func = phb_update_io
        # identify served workflows that ought to be tested
        wf_path2name = phb_get_wf_display_name(dockstore_path)
        exempted_tasks = {'version_capture'}
    else:  # set generic standards
        # do not define functions to parse and output these data
        pr_pop_func, io_update_func = None, None
        pr_template_path = pr_out_path
        wf_path2name = {}
        with open(pr_template_path, "w") as out:
            out.write("")

    logger.info(
        f"Compiling derived branch graph: {args.derived_branch}"
    )
    G_der, io_dict_der = compile_repo(repo, args.prefix, args.suffix)

    # compile additional and modififed repository nodes
    new_nodes = compile_affected_nodes(args.input, added_files, args.prefix)
    perturbed_nodes = compile_affected_nodes(args.input, modified_files, args.prefix)

    # compile and parse the source graph to obtain paths for removed nodes
    if source_branch == derived_branch:
        G_src, io_dict_src = G_der, io_dict_der
    else:
        logger.info(f"Checking out branch: {source_branch}")
        repo.git.checkout(source_branch)
        G_src, io_dict_src = compile_repo(repo, args.prefix, args.suffix)
    # create a translation mapping between tags
    src_tag2der_tag, tag2name_der, tag2name_src = translate_tags_and_names(
        io_dict_der, io_dict_src, wf_path2name
    )

    removed_nodes = compile_affected_nodes(args.input, removed_files, args.prefix)

    logger.info(f"Checking out branch: {args.derived_branch}")
    repo.git.checkout(args.derived_branch)
    # compile the paths affected by modified nodes
    perturbed_roots = obtain_graph_paths(
        G_der,
        new_nodes["wf"]
        + new_nodes["task"]
        + perturbed_nodes["wf"]
        + perturbed_nodes["task"],
        tag2name_der.keys(),
    )
    if removed_nodes["wf"] or removed_nodes["task"]:
        # need to translate between the source graph PATH to the derived graph PATH to obtain its tag
        perturbed_roots = {
            **perturbed_roots,
            **obtain_graph_paths(
                G_src, removed_nodes["wf"] + removed_nodes["task"], tag2name_src.keys()
            ),
        }

    removed_wf_set = set(removed_nodes["wf"])
    # create the testing dictionary by adding paths to workflows
    testing_dict = {}
    for k, v in sorted(perturbed_roots.items(), key=lambda x: x[0]):
        v = [x for x in v if x[1] not in exempted_tasks]
        if v:
            if k in io_dict_der:
                testing_dict[k[1]] = {"path": io_dict_der[k]["path"], "modified": v}
            # if it isn't here then there is an error potentially
            elif k in src_tag2der_tag:
                der_tag = src_tag2der_tag[k]
                testing_dict[der_tag[1]] = {
                    "path": io_dict_der[der_tag]["path"],
                    "modified": v,
                }
            elif k not in removed_wf_set:
                raise KeyError(f"Unknown error from modified node {k}")
    
    # only need the source branch if we are generating a PR or a test template
    clean_io_dict_der = {k[1]: v for k, v in io_dict_der.items()}
    clean_io_dict_src = {k[1]: v for k, v in io_dict_src.items()}
    io_diff_dict = parse_io_modifications(clean_io_dict_der, clean_io_dict_src)
    wf_diff_dict, task_diff_dict = parse_wf_task_modifications(
        new_nodes, perturbed_nodes, removed_nodes
    )

    if args.pull_request is not None:
        logger.info(
            f"Downloading PR #{args.pull_request} from {args.owner}/{args.repository}"
        )
        pr_str = download_pr(args.owner, args.repository, args.pull_request)
    else:
        logger.info(f"Populating new PR template from {pr_template_path}")
        with open(pr_template_path, "r") as pr_template:
            pr_str = pr_template.read()
    pr_str, testing_dict = pr_pop_func(
        wf_diff_dict, task_diff_dict, io_diff_dict, testing_dict, pr_str
    )

    # output files
    with open(io_path, "w") as out:
        out.write(json.dumps(clean_io_dict_der, indent=2))
    with open(f"{out_dir}{out_base}.testing.json", "w") as out:
        out.write(json.dumps(testing_dict, indent=2))
    with open(pr_out_path, "w") as out:
        out.write(pr_str)

    # update I/O if applicable
    if io_update_func:
        io_args = (
            clean_io_dict_der,
            i_tsv_path,
            o_tsv_path,
            wf_path2name,
            out_dir,
            out_base,
        )
        io_update_func(*io_args)

    repo.git.checkout(starting_branch)
    logger.info(f"Checking out branch: {starting_branch}")


if __name__ == "__main__":
    main()
    sys.exit(0)



#! /usr/bin/env python3

import logging
import networkx as nx
from collections import defaultdict

logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.INFO)


def get_root_nodes(G):
    """Get root nodes of the graph"""
    return [n for n, d in G.in_degree() if d == 0]


def get_leaf_nodes(G):
    """Get leaf nodes of the graph"""
    return [n for n, d in G.out_degree() if d == 0]


def obtain_graph_paths(G, nodes, root_nodes=None):
    """Obtain the paths from the root workflows through a node in the graph"""
    if not root_nodes:
        root_nodes = get_root_nodes(G)
    perturbed_roots = defaultdict(list)
    for node in nodes:
        for root in root_nodes:
            root_paths = list(nx.all_simple_paths(G, source=root, target=node))
            if root_paths:
                perturbed_roots[root].append(node)
                logger.info(f"{root[1]} affected by {node[1]}")
                continue
    return perturbed_roots



#! /usr/bin/env python3

import os
import re
import WDL
import logging
import networkx as nx

logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.INFO)


def parse_expr_apply(expr, parsed_expr):
    """Parse a WDL function applied to an expression"""
    # extract the function and prepare a dictionary for argument population
    apply_dict = {"func": expr.function_name, "args": []}
    for arg in expr.arguments:
        apply_dict["args"].append(parse_expr(arg, []))
    parsed_expr.append(apply_dict)
    return parsed_expr


def parse_expr_array(expr, parsed_expr):
    parsed_expr.append(expr.literal.json)
    return parsed_expr


def parse_expr_get(expr, parsed_expr):
    parsed_expr = parse_expr(expr.expr, parsed_expr)
    return parsed_expr


def parse_expr_ident(expr, parsed_expr):
    parsed_expr.append(str(expr.name))
    return parsed_expr


def parse_expr_string(expr, parsed_expr):
    """Extract a Python str from a WDL String while removing extraneous quotes"""
    deconstruction = "".join(
        [x for x in expr.parts if x.replace("'", "").replace('"', "")]
    )
    parsed_expr.append(deconstruction)
    return parsed_expr


def parse_expr_value(expr, parsed_expr):
    parsed_expr.append(expr.value)
    return parsed_expr


def parse_expr_map(expr, parsed_expr):
    for k, v in expr.items:
        parsed_expr.append([parse_expr(k, []), parse_expr(v, [])])
    return parsed_expr


def parse_expr(expr, parsed_expr):
    """Recrusively parse expression types"""
    if isinstance(expr, WDL.Expr.Apply):
        parsed_expr = parse_expr_apply(expr, parsed_expr)
    elif isinstance(expr, WDL.Expr.Get):
        parsed_expr = parse_expr_get(expr, parsed_expr)
    elif isinstance(expr, WDL.Expr.Ident):
        parsed_expr = parse_expr_ident(expr, parsed_expr)
    elif isinstance(expr, WDL.Expr.String):
        parsed_expr = parse_expr_string(expr, parsed_expr)
    elif isinstance(expr, WDL.Expr.Array):
        parsed_expr = parse_expr_array(expr, parsed_expr)
    elif isinstance(expr, (WDL.Expr.Int, WDL.Expr.Boolean, WDL.Expr.Float)):
        parsed_expr = parse_expr_value(expr, parsed_expr)
    elif isinstance(expr, (WDL.Expr.Map)):
        parsed_expr = parse_expr_map(expr, parsed_expr)
    else:
        # expression parsing function needs to be added
        raise AttributeError(f"unaccounted expression type: {type(expr)}; {str(expr)}")
    return parsed_expr


def parse_call(G, node, io_dict={}, source_callee=None, source_node=None):
    name2tag = {}

    # acquire the inputs delivered to the call
    callee = node.callee
    callee_tag = (
        callee.digest,
        callee.name,
    )
    # acquire the source information from the call
    if isinstance(callee, (WDL.Tree.Task, WDL.Tree.Workflow)):
        G.add_edge(source_callee, callee_tag)
        if source_node:
            derived_source = source_node + "." + node.name
        else:
            derived_source = node.name
        name2tag[derived_source] = callee_tag
        io_dict[derived_source] = {
            "callee": callee_tag,
            "inputs": populate_inputs(callee.inputs),
            "outputs": {x.name: str(x.type) for x in callee.outputs},
        }
    else:
        raise AttributeError(f"unexpected call type {type(callee)}")
    return G, io_dict, name2tag


def parse_scatter_condn(
    G, node, wf_tag, io_dict={}, source_callee=None, source_node=None
):
    """Parse a scatter or conditional expression and append to graph and io_dict as tasks/wfs appear"""
    in2expr = {}
    name2tag = {}
    for node in node.body:
        if isinstance(node, WDL.Tree.Call):
            if source_node:
                derived_source = source_node + "." + node.name
            else:
                derived_source = node.name
            call_inputs = link_inputs(node.inputs)
            for k, v in call_inputs.items():
                if v not in in2expr:
                    in2expr[v] = []
                in2expr[v].append(
                    (
                        derived_source,
                        k,
                    )
                )
            G, io_dict, t_name2tag = parse_call(
                G, node, io_dict, source_callee=source_callee, source_node=source_node
            )
            name2tag = {**name2tag, **t_name2tag}
        elif isinstance(node, (WDL.Tree.Conditional, WDL.Tree.Scatter)):
            G, io_dict, t_in2expr, t_name2tag = parse_scatter_condn(
                G,
                node,
                wf_tag,
                io_dict,
                source_callee=source_callee,
                source_node=source_node,
            )
            in2expr = combine_in2expr(t_in2expr, in2expr)
            name2tag = {**name2tag, **t_name2tag}
        # ignore variable definitions, raise an error otherwise
        elif not isinstance(node, WDL.Tree.Decl):
            raise AttributeError(
                f"unexpected node type {type(node)} in scatter/conditional body"
            )
    return G, io_dict, in2expr, name2tag


def link_inputs(inputs):
    """Link the inputs dictionary with the name and default values"""
    inputs_dict = {}
    for i, v in inputs.items():
        if isinstance(v, WDL.Expr.Get):
            data = [x for x in parse_expr(v.expr, [])]
            # ignore Arrays/Maps
            if len(data) == 1:
                inputs_dict[i] = data[0]

    return inputs_dict


def combine_in2expr(t_in2expr, in2expr):
    """Populate a dictionary relating workflows to inputs and expressions"""
    for k, v in t_in2expr.items():
        if k not in in2expr:
            in2expr[k] = []
        in2expr[k] += v
    return in2expr


def parse_wf(G, wf, wf_tag):
    """Obtain the 1st order data structures (I/O, call namespaces, and call input expressions) from a workflow"""
    io_dict = {}
    name2tag = {}
    in2expr = {}
    # parse the body of a workflow and compile its calls
    for node in wf.body:
        if isinstance(node, WDL.Tree.Call):
            derived_source = node.name
            call_inputs = link_inputs(node.inputs)
            for k, v in call_inputs.items():
                if v not in in2expr:
                    in2expr[v] = []
                in2expr[v].append(
                    (
                        derived_source,
                        k,
                    )
                )
            G, io_dict, t_name2tag = parse_call(G, node, io_dict, source_callee=wf_tag)
            name2tag = {**name2tag, **t_name2tag}
        elif isinstance(node, (WDL.Tree.Conditional, WDL.Tree.Scatter)):
            G, io_dict, t_in2expr, t_name2tag = parse_scatter_condn(
                G, node, wf_tag, io_dict, source_callee=wf_tag
            )
            in2expr = combine_in2expr(t_in2expr, in2expr)
            name2tag = {**name2tag, **t_name2tag}
        elif isinstance(node, WDL.Tree.Decl):
            continue  # these are just variable definitions, no action needed
        else:
            raise AttributeError(f"unexpected node type {type(node)} in workflow body")

    return G, io_dict, in2expr, name2tag


def populate_inputs(inputs):
    """Populate the inputs dictionary with the name, type, and default values"""
    inputs_dict = {}
    for i in inputs:
        inputs_dict[i.name] = {"type": str(i.type), "default": None}
        # if there is an expression indicating a default is set
        if i.expr:
            data = [x for x in parse_expr(i.expr, [])]
            # only maintain the list type for Arrays/Maps
            if len(data) == 1:
                inputs_dict[i.name]["default"] = data[0]
            else:
                inputs_dict[i.name]["default"] = data

    return inputs_dict


def compile_workflow_data_structures(workflow_file, G):
    """Import a WDL workflow, return its dependency graph, and I/O"""
    dep_io_dict = {}
    try:
        wdl = WDL.load(workflow_file)
    except Exception as e:
        raise RuntimeError(f"WDL parsing errors raised from {workflow_file}: {e}")

    # if this is a workflow, parse the workflow
    if wdl.tasks:
        raise AttributeError(
            f"Anticipated workflow WDL file {workflow_file} contains tasks"
        )
    elif wdl.workflow:
        wf_tag = (
            wdl.workflow.digest,
            wdl.workflow.name,
        )
        logger.debug(f"Parsing {wdl.workflow.name}")
        G, dep_io_dict, wf_in2expr, name2tag = parse_wf(G, wdl.workflow, wf_tag)
        wf_io_dict = {"inputs": {}, "outputs": {}}
        if wdl.workflow.inputs:
            wf_io_dict["inputs"] = populate_inputs(wdl.workflow.inputs)
        if wdl.workflow.outputs:
            wf_io_dict["outputs"] = WDL.values_to_json(wdl.workflow.effective_outputs)
        available_inputs = set(WDL.values_to_json(wdl.workflow.available_inputs))
        for dep, io_dict in dep_io_dict.items():
            for k, v in io_dict["inputs"].items():
                dep_input = f"{dep}.{k}"
                if dep_input in available_inputs:
                    wf_io_dict["inputs"][dep_input] = v
        if wf_tag not in G.nodes:
            G.add_node(wf_tag)
    else:
        raise AttributeError(
            f"Anticipated workflow WDL file {workflow_file} does not contain a workflow"
        )

    return G, wf_tag, wf_in2expr, dep_io_dict, name2tag, wf_io_dict


def extract_wdl_nodes(wdl_file):
    """Parse a WDL file and obtain the tasks/workflows contained"""
    node_names = []
    logger.debug(f"Parsing WDL file {wdl_file}")
    wdl = WDL.load(wdl_file)
    if wdl.workflow:
        wdl_tag = (
            wdl.workflow.digest,
            wdl.workflow.name,
        )
        node_names.append(wdl_tag)
    elif wdl.tasks:
        for task in wdl.tasks:
            node_names.append(
                (
                    task.digest,
                    task.name,
                )
            )
    else:
        raise AttributeError(
            f"Anticipated WDL file {wdl_file} does not contain a workflow or task"
        )
    return node_names


def float_default_from_dep_io_dict(i, linked_expressions, dep_io_dict):
    """Float defaults up to the top for optional inputs w/o a default and a single determined linked value"""
    defaults = set()
    for node_call, node_input in linked_expressions:
        if node_call not in dep_io_dict:
            raise KeyError(f"Node {node_call} not found in dependency I/O dictionary")
        elif node_input not in dep_io_dict[node_call]["inputs"]:
            raise KeyError(f"Input {node_input} not found in node {node_call} inputs")
        t_default = dep_io_dict[node_call]["inputs"][node_input]["default"]
        defaults.add(t_default)
    # if there is a single default value to link, float it
    if len(defaults) == 1:
        default = list(defaults)[0]
    else:
        default = None
        logger.debug(f"input: {i} is not used in imported modules")

    return default


def float_first_order_defaults(inputs_dict, dep_io_dict, in2expr):
    """Extract default input data from dependencies and apply to source input dictionary"""
    for i, input_metadata in inputs_dict.items():
        # this is the source WDL
        if i.count(".") == 0:
            if "?" in input_metadata["type"]:
                # if there isn't a default
                if input_metadata["default"] is None:
                    # if there is a downstream link
                    if i in in2expr:
                        input_metadata["default"] = float_default_from_dep_io_dict(
                            i, in2expr[i], dep_io_dict
                        )

    return inputs_dict


def float_nested_wf_defaults(wf2io, wf2in2expr, wf2name2tag):
    """Extract default input data from dependencies and apply to source input dictionary"""

    # initialize an iterative default float loop
    default_floated = True
    # continue as long as a default has been floated in the previous iteration
    while default_floated:
        default_floated = False
        for src_wf, io_dict in wf2io.items():
            for i, input_metadata in io_dict["inputs"].items():
                # if this is an optional input
                if "?" in input_metadata["type"]:
                    # if it doesn't have a default
                    if input_metadata["default"] is None:
                        # if this is a nested input
                        if i.count(".") == 0:
                            # if there is an expression to float from
                            if i in wf2in2expr[src_wf]:
                                input_expressions = wf2in2expr[src_wf][i]
                                defaults = set()
                                for node_call, node_input in input_expressions:
                                    node_tag = wf2name2tag[src_wf][node_call]
                                    if node_tag in wf2io:
                                        defaults.add(
                                            wf2io[node_tag]["inputs"][node_input][
                                                "default"
                                            ]
                                        )
                                if len(defaults) == 1:
                                    default = list(defaults)[0]
                                    if default is not None:
                                        input_metadata["default"] = list(defaults)[0]
                                        default_floated = True

    # set nested defaults from their wf source now that they are comprehensively populated
    for src_wf, io_dict in wf2io.items():
        for i, input_metadata in io_dict["inputs"].items():
            if "?" in input_metadata["type"]:
                if input_metadata["default"] is None:
                    if i.count(".") == 1:
                        node_data = re.search(r"(.*)\.([^\.]+$)", i)
                        node_call = node_data.group(1)
                        node_input = node_data.group(2)
                        node_tag = wf2name2tag[src_wf][node_call]
                        # if it is a workflow, grab its default
                        if node_tag in wf2io:
                            input_metadata["default"] = wf2io[node_tag]["inputs"][
                                node_input
                            ]["default"]
    return wf2io


def compile_io(wf2io, wf2dependency_io, wf2namespace2tag, wf2inputs2expressions):

    # default floating could probably be done more efficiently with graph traversal
    for wf_tag, io in wf2io.items():
        io["inputs"] = float_first_order_defaults(
            io["inputs"], wf2dependency_io[wf_tag], wf2inputs2expressions[wf_tag]
        )

    wf2io = float_nested_wf_defaults(wf2io, wf2inputs2expressions, wf2namespace2tag)

    for io_dict in wf2io.values():
        io_dict["inputs"] = {
            k: v
            for k, v in sorted(io_dict["inputs"].items(), key=lambda x: x[0].count("."))
        }

    return wf2io


def compile_repo(repo, wf_prefix="wf_", wf_suffix=".wdl"):
    """Acquire all workflows and compile the repository's dependency graph"""
    # get the base repository path
    repo_dir_path = os.path.abspath(repo.working_tree_dir)

    # acquire all workflows in the repository
    workflows = {}
    for file in repo.git.ls_files().splitlines():
        wf_full_path = os.path.join(repo_dir_path, file)
        if os.path.basename(file).startswith(wf_prefix) and os.path.basename(
            file
        ).endswith(wf_suffix):
            workflows[wf_full_path] = file

    # initialize the graph
    G = nx.DiGraph()
    # init the io dict
    wf2io, wf2dependency_io, wf2namespace2tag, wf2inputs2expressions = {}, {}, {}, {}
    for raw_wf_path, repo_path in workflows.items():
        G, wf_tag, wf_in2expr, dep_io_dict, name2tag, wf_io_dict = (
            compile_workflow_data_structures(raw_wf_path, G)
        )
        wf2io[wf_tag] = wf_io_dict
        wf2io[wf_tag]["path"] = repo_path
        wf2dependency_io[wf_tag] = dep_io_dict
        wf2namespace2tag[wf_tag] = name2tag
        wf2inputs2expressions[wf_tag] = wf_in2expr

    io = compile_io(wf2io, wf2dependency_io, wf2namespace2tag, wf2inputs2expressions)

    return G, {k: v for k, v in sorted(io.items(), key=lambda x: x[0])}
